#!/usr/bin/env python

import torch, torchvision
from torch import nn

######################################################################

def _flatparam(model, whole, already = [], offset = 0):
    for v in model._parameters:
        p = model._parameters[v]
        e = p.numel()
        s = p.size()
        model._parameters[v] = whole[offset:offset+e].view(s)
        with torch.no_grad():
            model._parameters[v].copy_(p)
        offset += e
    already.append(model)
    for m in model.modules():
        if m not in already:
            offset = _flatparam(m, whole, already, offset)
    return offset

def flatparam(model):
    n = sum(p.numel() for p in model.parameters())
    whole = next(model.parameters()).new(n) # Get same device and dtype
    whole.requires_grad_()
    _flatparam(model, whole)
    model.parameters = lambda: iter([ whole ])

######################################################################

model = nn.Sequential(
    nn.Linear(2, 4),
    nn.ReLU(),
    nn.Sequential(
        nn.Linear(4, 4),
        nn.ReLU(),
        nn.Linear(4, 2)
    )
)

######################################################################

print('Before:')
for p in model.parameters():
    print(p.size(), p.storage().size())

flatparam(model)

print('After:')
for p in model.parameters():
    print(p.size(), p.storage().size())

######################################################################

print('Check:')

input = torch.rand(100, 2)
targets = torch.rand(100, 2)
optimizer = torch.optim.SGD(model.parameters(), lr = 1e-2)
mse = nn.MSELoss()

for e in range(10):
    output = model(input)
    loss = mse(output, targets)
    print(e, loss.item())
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()
